//Description: loads geometry from a named node within an OpenFlight file
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> a source and node are specified
//<li> default source is called "fltSource"
//<li> default node is the name of the fltObject
//<li> desired node is found under source, removed, and given to the fltObject
//</ul>
//
// Category: Geometry
// Author: Alex Hill
// Revision: 11/01/01
//
#include <Performer/pfdu.h>
#include <Performer/pf/pfGroup.h>
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "ygNodeDB.h"
#include "ygPFObjectCache.h"
#include "fltObject.h"

extern "C" ygNode* construct_fltObject(const char* name,bool master) { return new fltObject(name,master); }

struct _fltObjectPrivateData
    {
    ygString filename;
    ygString sourceNode;
    ygString sourceName;
    int sourceMode;
    pfNode *obj;
    bool clean;
    };


fltObject::fltObject(const char* name,bool master) : ygGeometry(name,master)
    {
    setClassName("fltObject",true);
    p_ = new struct _fltObjectPrivateData;
    p_->obj = NULL;
    p_->clean = false;
    p_->sourceMode = 0;
    p_->sourceName = this->name();
    addNetMessage("file",&p_->filename,YG_NET_STRING);
    addNetMessage("file",&p_->clean,YG_NET_BOOL);
    addNetMessage("source",&p_->sourceNode,YG_NET_STRING);
    addNetMessage("source",&p_->sourceName,YG_NET_STRING);
    addNetMessage("source",&p_->sourceMode,YG_NET_INT);
    addNetMessageMap("source","none",0);
    addNetMessageMap("source","parent",1);
    addNetMessageMap("source","child",2);
    dontSendNetMessage("source");
    pfdInitConverter("flt");
    pfdConverterMode("flt",1,FALSE);
    }


fltObject::~fltObject(void)
    {
    delete p_;
    }


void fltObject::reset(void)
    {
    ygGeometry::reset();
    }


void fltObject::message(const ygMessage& msg)
    {
    //load the given case sensitive filename
    if (msg == "file")
        {
        if (p_->filename.length() > 0)
            loadFile(p_->filename,p_->clean);
        }
    //extract geometry with a given name from the default source node
    else if (msg == "node")
        {
        ygString fltSource("fltSource");
        if (msg.numArgs() == 0)
            extractNode(fltSource,name(),0);
        else if (msg.numArgs() == 1)
            extractNode(fltSource,msg.stringArg(0),2);
        else if (msg.numArgs() == 2)
            {
            if (msg.stringArg(1) == "parent")
                extractNode(fltSource,msg.stringArg(0),1);
            else
                extractNode(fltSource,msg.stringArg(0),0);
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //extract geometry with a given name from a specific source node
    else if (msg == "source")
        {
        if (msg.numArgs() == 2)
            p_->sourceMode = 2;
        extractNode(p_->sourceNode,p_->sourceName,p_->sourceMode);
        }
    else
        ygGeometry::message(msg);
    }


void fltObject::loadFile(const ygString& file,const bool& clean)
    {
    if (p_->obj)
        pfnode()->removeChild(p_->obj);
    if (!clean)
        pfdConverterMode("flt",2,FALSE);
    p_->obj = ygPFObjectCache::objectReference(file.c_str());
    if (p_->obj)
        {
        p_->filename = file;
        pfnode()->addChild(p_->obj);
        p_->obj->setTravMask(PFTRAV_ISECT, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
        p_->obj->setTravMask(PFTRAV_DRAW, 0xffffffff, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
        }
    }


void fltObject::extractNode(const ygString& source, const ygString& node, const int& mode)
        {
        if (p_->obj) 
                pfnode()->removeChild(p_->obj);
        ygNode *ygSource = ygNodeDB::find(source.c_str());
        if (ygSource)
        {
            pfGroup* pointer;
                pointer = (pfGroup*) ygSource->pfnode()->find(node.c_str(), pfGroup::getClassType());
            if (mode == 0)
                p_->obj = pointer;
            else if (mode == 1)
                p_->obj = pointer->getParent(0);
            else if (mode == 2)
                p_->obj = pointer->getChild(0);
        if (p_->obj)
            {
            pfGroup *parent = (pfGroup*) p_->obj->getParent(0);
            parent->removeChild(p_->obj);
            }
        else
            printf("fltObject: ERROR - node %s not found\n",node.c_str());
        }
        else
        printf("fltObject: ERROR - source node %s not found\n",source.c_str());
        if (p_->obj)
        {
        p_->obj->setTravMask(PFTRAV_ISECT, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
        p_->obj->setTravMask(PFTRAV_DRAW, 0xffffffff, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
        pfnode()->addChild(p_->obj);
        p_->sourceNode = source;
        p_->sourceName = node;
        p_->sourceMode = mode;
        netMessageChanged("source");
        }
        }
