//Description: displays a live video stream
//
//<b>notes:</b>
//<ul>
//<li> all parent node messages are inherited
//<li> see reset method for default settings
//<li> this node is currently supported only on LINUX machines
//<li> uses the JPEGDecompress C++ class to decompress jpeg data
//<li> the image is received through network via a non-blocking UDP socket
//<li> a video source of 320x240 in IYUV format is recommended
//<ul> Requirements:
//<li> a Windows PC
//<li> a Web Camera, connected to the Windows PC
//<li> The <a href="http://www.evl.uic.edu/eleni/yg/winAvatarCam.html">WinAvatarCam Application</a> for capturing and streaming video.
//<li> to heck whether UDP data are actually received in the Linux machine you can use <a href="http://sourceforge.net/projects/xosview/">xosview </a>
//</ul>
//</ul>
//
//
// Category: Attributes
// Author: Kalle Jalkanen and Jonas Westling, Interactive Institute Ume, Sweden
//           04/26/02
//           08/10/03 Helen-Nicole Kostis - added documentation
// Revision: 02/20/06 Alex Hill - stripped down and reconfigured to use existing geometry nodes
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfTexture.h>
#include <QUANTA/QUANTAnet_udp_c.hxx>
#include <QUANTA/QUANTA.hxx>
#include "JPEGDecompress.h"
#include "ygNetMessages.h"
#include "videoTexture.h"

using namespace std;

extern "C" ygNode* construct_videoTexture(const char* name,bool master) { return new videoTexture(name,master); }

struct _videoTexturePrivateData
    {
    JPEGDecompress djpeg;        
    bool active;
    pfTexture* texture;      
    QUANTAnet_udp_c *socket;    
    int port;                
    unsigned char packet[65536];    
    unsigned char jpegdata[65536];
    unsigned char* buffer;
    pfMatrix* mat;
    int nrOfPackets;
    int byteCounter;
    int lastW;
    int lastH;
    int pow2W;
    int pow2H;
    };


videoTexture::videoTexture(const char* name, bool master) : ygNode(name, master)
    {
    setClassName("videoTexture",true); 
    p_ = new struct _videoTexturePrivateData;
    p_->active = false;
    addNetMessage("play", &p_->active, YG_NET_BOOL_NEGATE);    
    p_->port = -1;
    p_->socket = NULL;
    if ( p_->djpeg.init() > 0 )
        cerr << "videoTexture: JPEG decompressor failed to initialize properly" << endl;
    p_->nrOfPackets = -1;
    p_->byteCounter = 0;
    p_->buffer = (unsigned char*)pfMalloc(512*256*3,pfGetSharedArena());
    memset(p_->buffer,0,512*256*3);
    p_->texture = new pfTexture;
    p_->texture->setFormat(PFTEX_IMAGE_FORMAT, PFTEX_RGB);
    p_->texture->setFormat(PFTEX_INTERNAL_FORMAT, PFTEX_RGB_8);
    p_->texture->setImage((uint*)p_->buffer,3,512,256,1);
    p_->mat = (pfMatrix*) pfMalloc(sizeof(pfMatrix), pfGetSharedArena());
    p_->mat->makeScale(320/512.0,240/256.0,1);
    p_->lastW = 320;
    p_->lastH = 240;
    p_->pow2W = 512;
    p_->pow2H = 256;
    }


videoTexture::~videoTexture()
    {   
    if (p_->socket) 
        delete p_->socket;
    pfDelete(p_->mat);
    if (p_->buffer)
        pfDelete(p_->buffer);
    delete p_;
    }


void videoTexture::reset()
    {
    ygNode::reset();
    }


void videoTexture::message(const ygMessage& msg)
    {
    //start the video
    if (msg == "play") 
        {
        return;
        } 
    //stop the video
    else if (msg == "stop")
        {
        p_->active = false;
        netMessageChanged("play");
        }
    //set the port to receive JPEG stream
    else if (msg == "port") 
        {
        if (!p_->socket) 
            {
            p_->port = (int)msg.intArg(0);
            p_->socket = new QUANTAnet_udp_c(); 
            p_->socket->init(p_->port);
            }
        } 
    else 
        ygNode::message(msg);
    }


#define MAXPAYLOAD 1400
#define HEADERSIZE 7

int videoTexture::readImageData()
    {
    if (!p_->socket)
        {
        return 0;
        }
    int bytesRecv = p_->socket->receive((char *)p_->packet, 65535, QUANTAnet_udp_c::NON_BLOCKING);   
    if (bytesRecv > 0) 
        {
        int imageNr = (int)p_->packet[0];   
        int picPartNr = (int)p_->packet[1];
        int isLast = (int)p_->packet[2];
        
        if (picPartNr == 0) 
            {
            if (p_->nrOfPackets != -1) 
                {
                cerr << "videoTexture: dropped UDP packets" << endl;
                }
            p_->byteCounter = 0;
            p_->nrOfPackets = 0;    
            } 
        else 
            p_->nrOfPackets++;
        
        memcpy(p_->jpegdata + picPartNr*MAXPAYLOAD, p_->packet + HEADERSIZE, bytesRecv - HEADERSIZE);    
        p_->byteCounter += (bytesRecv - HEADERSIZE);    
        
        if (isLast && (picPartNr==p_->nrOfPackets)) 
            {
            p_->nrOfPackets = -1;
            return p_->byteCounter;          
            } 
        else 
            {
            if (isLast) 
                {        
                cerr << "videoTexture: lost part of the image: got (";
                cerr << p_->nrOfPackets+1 << "/" << picPartNr+1 << ") packets"<<endl;
                p_->nrOfPackets = -1;                    
                }
            }   
        } 
    else if (bytesRecv != QUANTAnet_udp_c::NON_BLOCKING_HAS_NO_DATA) 
        {
        cerr << "videoTexture: QUANTA error - UDP socket receive failed (code=" << bytesRecv << ")" << endl;    
        delete p_->socket;
        p_->socket = NULL;    
        p_->nrOfPackets = -1;
        p_->byteCounter = 0;
        }
    return 0;
    }


int log2(int x) 
    {
    int a = -1;
    while (x > 0) 
        {
        x>>=1;
        a++;
        }
    return a;
    }


void videoTexture::app(void)
    {
    if (isChanged(YG_CHANGE_GEOMETRY | YG_CHANGE_TEXTURE))
        replaceTextures(pfnode());
    if (p_->active) 
        {    
        int numbytes = 0;
        numbytes = readImageData();
        if (numbytes > 0) 
            {
            p_->djpeg.decompress(p_->jpegdata, numbytes);
            int h,w;
            w = p_->djpeg.getWidth();
            h = p_->djpeg.getHeight();
            if (w != p_->lastW || h != p_->lastH)
                {
                char log2w = (char)log2(w);
                char log2h = (char)log2(h);
                if (w > (1<<log2w)) log2w++;
                if (h > (1<<log2h)) log2h++;
                
                p_->pow2W = 1 << log2w;
                p_->pow2H = 1 << log2h;
                pfDelete(p_->buffer);
                p_->buffer = (unsigned char*)pfMalloc(p_->pow2W*p_->pow2H*3,pfGetSharedArena());
                p_->mat->makeScale(w/(float)p_->pow2W,h/(float)p_->pow2H,1);
                p_->lastW = w;
                p_->lastH = h;
                }
            if (w == p_->pow2W && h == p_->pow2H)
                memcpy(p_->buffer, p_->djpeg.data(), w*h);
            else
                copySubImage(p_->djpeg.data(),w,h,3);            
            p_->texture->setFormat(PFTEX_IMAGE_FORMAT, PFTEX_RGB);
            p_->texture->setFormat(PFTEX_INTERNAL_FORMAT, PFTEX_RGB_8);
            p_->texture->setImage((uint*)p_->buffer,3,p_->pow2W,p_->pow2H,1);
            }       
        }
    ygNode::app();
    }


void videoTexture::copySubImage(unsigned char *img, int w, int h, int ncolc)
    {
    if (img && p_->buffer) 
        {
        int x,y,i;
        int od, os;
        
        for (y=0; y<h; ++y) 
            {
            od = ncolc * p_->pow2W * y;
            os = ncolc * w * y;               
            for (x=0; x<w; ++x) 
                {
                for (i=0; i<ncolc; ++i) 
                    {
                    p_->buffer[od+i] = img[os+i];
                    }
                od += ncolc;
                os += ncolc;
                }
            }
        }
    }


void videoTexture::replaceTextures(pfNode * node)
    {
    int i;
    if (node->isOfType(pfGeode::getClassType()))
        {
        pfGeode *geode = (pfGeode *)node;
        for (i=0; i < geode->getNumGSets(); i++)
            {
            pfGeoSet* gset = geode->getGSet(i);
            pfGeoState* gstate = gset->getGState();
            pfTexture* texture;
            texture = (pfTexture*)gstate->getAttr(PFSTATE_TEXTURE);
            if (texture)
                {
                gstate->setAttr(PFSTATE_TEXTURE, p_->texture);
                gstate->setMode(PFSTATE_ENTEXMAT, PF_ON);
                gstate->setAttr(PFSTATE_TEXMAT, p_->mat);
                }
            }
        }
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup *group = (pfGroup *)node;
        for (i=0; i < group->getNumChildren(); i++)
            replaceTextures(group->getChild(i));
        }
    }



