// Description: toggles the child for node processing and Performer traversal
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the alias: switch
//<li> creates a switch using a Performer pfSwitch
//<li> this node is frequently used to toggle geometry on and off
//</ul>
//
// Category: Selection
// Author: Dave Pape
//           11/01/01
//           06/28/05 Alex Hill - changed netOn variable to netOff for saving and printout
// Revision: 02/01/06 Alex Hill - changed netOn to netOff and moved event generation out of app function
//
#include <Performer/pf/pfSwitch.h>
#include "ygNetMessages.h"
#include "ygSwitch.h"

extern "C" ygNode* construct_ygSwitch(const char* name,bool master) { return new ygSwitch(name,master); }

struct _ygSwitchPrivateData
    {
    pfSwitch * sw;
    bool prevOff;
    bool netOff;
    };


ygSwitch::ygSwitch(const char* name,bool master) : ygNode(name,master,false)
    {
    setClassName("ygSwitch",true);
    p_ = new struct _ygSwitchPrivateData;
    //create pfSwitch for the pfnode
    p_->sw = new pfSwitch;
    p_->netOff = false;
    setPfNode(p_->sw);
    addNetMessage("off",&p_->netOff,YG_NET_BOOL_NEGATE);
    dontSendNetMessage("off");
    addEvent("switchOn");
    addEvent("switchOff");
    }


ygSwitch::~ygSwitch(void)
    {
    delete p_;
    }


void ygSwitch::reset(void)
    {
    //if net master then set the switch to on
    if (isNetMaster())
        {
        on();
        p_->prevOff = false;
        }
    ygNode::reset();
    }


/***** Message functions *****/

void ygSwitch::message(const ygMessage& msg)
    {
    //set the switch to on
    if (msg == "on")
        {
        on();
        }
    //set the switch to off
    else if (msg == "off")
        {
        if (p_->netOff)
            off();
        else
            on();
        }
    //toggle the state of the switch
    else if (msg == "toggle")
        {
        toggle();
        }
    else
        ygNode::message(msg);
    }


void ygSwitch::appTraverse(void)
    {
    //start timing
    timingBegin();
    //execute app method
    app();
    //if switch is on
    if (isOn())
        {
        //call appTraverse recursively on all children
        for (int i=0; i < numChildren(); i++)
            child(i)->appTraverse();
        }
    //end timing
    timingEnd();
    }


/*****  Switch functions *****/

bool ygSwitch::isOn(void) const
    {
    return (p_->sw->getVal() == PFSWITCH_ON);
    }


void ygSwitch::on(void)
    {
    if (!isOn())
        {
        p_->sw->setVal(PFSWITCH_ON);
        p_->netOff = false;
        p_->prevOff = true;
        netMessageChanged("off");
        //the switch has changed to on
        eventOccurred("switchOn");
        }
    }


void ygSwitch::off(void)
    {
    if (isOn())
        {
        p_->sw->setVal(PFSWITCH_OFF);
        p_->netOff = true;
        p_->prevOff = false;
        netMessageChanged("off");
        //the switch has changed to off
        eventOccurred("switchOff");
        }
    }


void ygSwitch::toggle(void)
    {
    if (isOn())
        off();
    else
        on();
    }
